/*
 * version.h
 *
 * Version number header file for CallGen323
 *
 * Copyright (c) 1993-2000 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): ______________________________________.
 *
 *
 * Library dependencies:
 *
 *   pwlib: v1.4.8   CVS tag: v1_4_8
 *   openh323: v1.11.3   CVS tag: v1_11_3
 */

#ifndef _CallGen323_VERSION_H
#define _CallGen323_VERSION_H

#define MAJOR_VERSION 1
#define MINOR_VERSION 2
#define BUILD_TYPE    ReleaseCode
#define BUILD_NUMBER 6


#endif  // _CallGen323_VERSION_H


// End of File ///////////////////////////////////////////////////////////////
